/***************************************************************************
 *   Copyright (C) 2004 by Mauricio Oliveira Carneiro (PROCC - FIOCRUZ)    *
 *   carneiro@fiocruz.br                                                   *
 *   http://www.procc.fiocruz.br/                                          * 
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "evolmodel.h"
#include "strain.h"
#include "clockgen.h"
#include "host.h"


#define DEBUG

int round (double x)
{
	return (int) (x + 0.5);
}

void DetermineInvariableSites (double rate, int size)
{
	int n = round(rate * size),
	    k;
	while (n--) {
		k = rand()%size;
		while (g_InvariableSites[k]) k = (k+1) % size;
		g_InvariableSites[k] = 1;
	}
}


int main (int argc, char *argv[])
{
	int nGenerations,
	    iParseOk,
	    i,
	    bWriteDescfile,
	    bInitEvolutionaryModel,
		bRandomFirstSequence,
		random_seed = time(NULL);
	double rateA = 0.25,
	       rateC = 0.25,
	       rateG = 0.25,
	       rateT = 0.25,
		   fInvariableSitesRate = 0.0;
	char filename[35];
	FILE *descfile;
	pStrain Original;
	EvolutionaryModel Model;
	pHost Host;
	
	/* parse the input */
	Original = (pStrain) malloc (sizeof(Strain));
	Original->iGeneration = 1;
	Original->iSeqCount = 0;
	Host = host_init(0.0);
	iParseOk = 0;
	bWriteDescfile = 0;
	bRandomFirstSequence = 1;
	bInitEvolutionaryModel = 1;
	strcpy(Original->Name, DEFAULT_TAXON_NAME ); /* default name for the taxa */

	for (i=1; i<argc; i++) {	
		if (!strcmp(argv[i], "-n")) { 
			iParseOk += 1;
			Original->nNucleotides = atoi(argv[++i]);
			Original->Nucleotides = (pSequence) malloc (Original->nNucleotides * sizeof(Sequence));
			g_InvariableSites = (char *) calloc (sizeof(char), atoi(argv[i]));
		}
		else if (!strcmp(argv[i], "-m")) {
			int j;
			iParseOk += 2;
			Original->nMutationsPerEvolution = atoi(argv[++i]);
		}
		else if (!strcmp(argv[i], "-g")) { 
			iParseOk += 4;
			nGenerations = atoi(argv[++i]);	
		}
		else if (!strcmp(argv[i], "-i")) { 
			iParseOk += 8;
			bRandomFirstSequence = 0;
			strain_read_fasta(Original, argv[++i]);
		}	
		else if (!strcmp(argv[i], "-s")) {
			if (strlen(argv[i+1]) >= 30) {
				fprintf(stderr, "Name is too long, must be 30 chars maximum");
				exit(1);
			}
			strcpy(Original->Name, argv[++i]);
		}
		else if (!strcmp(argv[i], "-em")) {
			bInitEvolutionaryModel = 0;
			if (!strcmp (argv[i+1], "JC")) {
				evolutionarymodel_init_JC(&Model);
				i++;
			}
			else if (!strcmp (argv[i+1], "K2P")) {
				evolutionarymodel_init_K2P(&Model, atof(argv[i+2]), atof(argv[i+3]));
				i+=3;
			}
			else if (!strcmp (argv[i+1], "REV")) {
				evolutionarymodel_init_REV(&Model, atof(argv[i+2]), atof(argv[i+3]), atof(argv[i+4]), 
				                                   atof(argv[i+5]), atof(argv[i+6]), atof(argv[i+7]));
				i+=7;
			}
		}
		else if (!strcmp(argv[i], "-f")) {
			rateA = atof(argv[++i]);
			rateC = atof(argv[++i]);
			rateG = atof(argv[++i]);
			rateT = atof(argv[++i]);
		}
		else if (!strcmp(argv[i], "-d")) 
			bWriteDescfile = 1;
		else if (!strcmp(argv[i], "-is"))
			fInvariableSitesRate = atof(argv[++i]);
		else if (!strcmp(argv[i], "-z"))
			random_seed = atof(argv[++i]);
		else if (!strcmp(argv[i], "-hi")) { 
			Host->Immunity = atof(argv[++i]);
		}	

	}

	/* Wrong Parameter format */
	if (iParseOk != 7 && iParseOk != 14) {
		fprintf(stderr, "Usage: clockgen [options]\n\n");
		fprintf(stderr, "OPTIONS:\n\n"); 
		
		fprintf(stderr, "-g number_of_generations [* REQUIRED *]\n");
		fprintf(stderr, "Number of generations to be simulated.\n\n");
		
		fprintf(stderr, "-m mutations_per_evolution [* REQUIRED *]\n");
		fprintf(stderr, "Number of mutations to occur on each evolutionary step.\n\n");
		
		fprintf(stderr, "-s strain_name\n");
		fprintf(stderr, "Name of the strains (indexes will be added in the output).\n");
		fprintf(stderr, "Example:\n");
		fprintf(stderr, " -s strain  => strain0x1, strain1x2, strain2x2, strain3x3, strain4x3, strain5x3, ...\n\n");
		
		fprintf(stderr, "-n nucleotides [* MUST USE ONE OF THESE *]\n");
		fprintf(stderr, "Number of nucleotides on the sequence to be generated. This option implies that"); 
		fprintf(stderr, "the program must generate a random first sequence. You must use one of the two options -n or -i.\n\n");
		
		fprintf(stderr, "-i first_sequence [* MUST USE ONE OF THESE *]\n");
		fprintf(stderr, "Fasta file with 1 strain to be used as the first strain in the simulation.");
		fprintf(stderr, "You must use one of the two options -n or -i.\n\n");
		
		fprintf(stderr, "-hi\n");
		fprintf(stderr, "Defines the host immunity factor [0-1] to this species, if not defined, 0 is assumed.\n\n");
		
		fprintf(stderr, "-d\n");
		fprintf(stderr, "Writes the description file with the name of the strain appended by .desc\n\n");
		
		fprintf(stderr, "-em Model [parameters]\n");
		fprintf(stderr, "Uses the evolutionary model described. If not specified, the JC model is used. The following models are implemented:\n");
		fprintf(stderr, "Model                     - Parameters (All parameters are decimal numbers)\n\n");
		fprintf(stderr, "JC  (Jukes-Cantor)        - No parameters.\n");
		fprintf(stderr, "K2P (Kimura 2 Parameters) - transitions_rate transversions_rate.\n");
		fprintf(stderr, "REV (General Reversible)  - AC AG AT CG CT GT.\n\n");
		
		fprintf(stderr, "-f rateA rateC rateG rateT\n");
		fprintf(stderr, "Specify the relative frequencies of the four nucleotides with decimal numbers representing the frequencies of A, C, G and T respectively. The frequencies MUST sum 1.0\n\n");
		
		fprintf(stderr, "-is rate\n");
		fprintf(stderr, "Specify the Invariable Sites rate. This proportion of sites will be kept unchanged through all simulation.\n\n");
		
		fprintf(stderr, "-z seed\n");
		fprintf(stderr, "Specify the seed for the random number generation.\n\n");
		exit(0);
	}

	/* Description file output */
	if (bWriteDescfile) {
		strcpy(filename, Original->Name);
		descfile = fopen (strcat(filename, ".desc"), "w");
		fprintf(descfile, "Generations: %d\n", nGenerations);
		fprintf(descfile, "Nucleotides: %d\n", Original->nNucleotides);
		fprintf(descfile, "Clock: %d\n", Original->nMutationsPerEvolution);
		fclose(descfile);
	}

	srand(random_seed);
	
	if (bInitEvolutionaryModel) 
		evolutionarymodel_init_JC(&Model);

	if (bRandomFirstSequence)
		strain_generate_random(Original, rateA, rateC, rateG, rateT);
	
	DetermineInvariableSites(fInvariableSitesRate, Original->nNucleotides);	
		
#ifdef DEBUG 
	printf("nGenerations : %d\n", nGenerations);
	printf("Host Immunity: %.2lf\n", Host->Immunity);
	printf("nMutations : %d\n", Original->nMutationsPerEvolution);
	printf("rateA: %lf\n", rateA);
	printf("rateC: %lf\n", rateC);
	printf("rateG: %lf\n", rateG);
	printf("rateT: %lf\n", rateT);
	printf("Invariables: ");
	for (i=0; i<Original->nNucleotides; i++)
		printf("%d ", g_InvariableSites[i]);
	printf("\n");
	printf("\n\n");
#endif
	/* Start Simulation */
	g_iSeqCount = 1;
	strain_evolve(Host, Model, Original, nGenerations, 1);
	return 0;
}
