/***************************************************************************
 *   Copyright (C) 2004 by Mauricio Oliveira Carneiro (PROCC - FIOCRUZ)    *
 *   carneiro@fiocruz.br                                                   *
 *   http://www.procc.fiocruz.br/                                          * 
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include "evolmodel.h"
#include "clockgen.h"
#include "strain.h"


char get_nucleotide (EvolutionaryModel model, char old) 
{
	int new;
	do 
		new = (rand()%100) + 1; 
	while (apply_model(model, old, new) == old);
	switch (new) { 
		case 0 : return 'A';
		case 1 : return 'C';
		case 2 : return 'G';
		default: return 'T';
	}
}

pStrain strain_init (int nNucleotides, int nMutationsPerEvolution, char *sName) 
{
	pStrain s = malloc (sizeof(Strain));
	s->nNucleotides = nNucleotides;
	s->nMutationsPerEvolution = nMutationsPerEvolution;
	s->Nucleotides = calloc (nNucleotides, sizeof(Sequence));
	s->iSeqCount = g_iSeqCount++;
	strcpy(s->Name, sName);
	return s;
}

void strain_generate_random (pStrain s, double rateA, double rateC, double rateG, double rateT) 
{
	int i,v;
	rateA *= 1000;
	rateC  = rateC*1000 + rateA;
	rateG  = rateG*1000 + rateC;
	rateT  = rateT*1000 + rateG;
	/* Attention to the fact that if the rates don't sum up to 1.0
	   we get a greater probability of having T's. */
	for (i=0; i < s->nNucleotides; i++) {
		v = rand()%1000;
		if (v <= rateA)      s->Nucleotides[i].Value = 'A';
		else if (v <= rateC) s->Nucleotides[i].Value = 'C';
		else if (v <= rateG) s->Nucleotides[i].Value = 'G';
		else                 s->Nucleotides[i].Value = 'T';
	}
	s->iGeneration = 1;
}

void strain_read_fasta (pStrain s, char *filename)
{
	int i;
	FILE *fasta_in;
	char buf[100];
	fasta_in = fopen(filename, "r");
	while (fgets(buf, 100, fasta_in)) {
		if (buf[0] == '>') {
			sscanf(buf, "%s", s->Name);
			if (s->Nucleotides) free(s->Nucleotides);
			s->nNucleotides = 0;
		}
		else if (buf[0] == '#') continue;
		else { 
			int j = s->nNucleotides,
			l = strlen(buf) - 1; /* chop the '\n' */ 
			s->nNucleotides += l;
			s->Nucleotides = (pSequence) realloc (s->Nucleotides, s->nNucleotides * sizeof(Sequence));
			for (i=0; i<l; i++) 
				s->Nucleotides[j+i].Value = buf[i];
		}
	}
}

void strain_show (pStrain s)
{
	int i;
	fprintf(stderr, "[%d]  -  ", s->iGeneration);
	for (i=0; i < s->nNucleotides; i++) 
		fprintf(stderr, "%c", s->Nucleotides[i].Value);
	fprintf(stderr, "\n");
}

void strain_output_fasta (pStrain s)
{
	int i;
	printf(">%s%dx%d\n", s->Name, s->iSeqCount, s->iGeneration);
	for (i=0; i < s->nNucleotides; i++) {
		printf("%c", s->Nucleotides[i].Value);
		if (!((i+1)%80)) printf("\n");
	}
	printf("\n");
}

void strain_delete (pStrain *s)
{
	free((*s)->Nucleotides);
	free(*s);
}


void strain_evolve (pHost h, EvolutionaryModel Model, pStrain s, int nGenerations, int iGeneration)
{
	int i,
	nMutations = s->nMutationsPerEvolution;
	pStrain s1,s2;
	char Mutations1[s->nNucleotides];  /* keeps track of all the sites wether they have mutated or not for first son*/
	char Mutations2[s->nNucleotides];  /* keeps track of all the sites wether they have mutated or not for second son*/
	
	/*	if (nGenerations == iGeneration)  // little quick and dirty hack to generate tests */ 
	strain_output_fasta(s);

	/* Lineage breakpoint */
	if (h->Immunity * 100 > rand() % 100) {
		strain_delete(&s);
		return;
	}
	
	if (iGeneration >= nGenerations) return;

	s1 = strain_init(s->nNucleotides, s->nMutationsPerEvolution, s->Name);
	s2 = strain_init(s->nNucleotides, s->nMutationsPerEvolution, s->Name);

	/* perfect copies to be mutated */
	for (i=0; i<s->nNucleotides; i++) {
		s2->Nucleotides[i].Value = s1->Nucleotides[i].Value = s->Nucleotides[i].Value;
	}

	s1->iGeneration = s2->iGeneration = iGeneration + 1;

	/* s->nNucleotides should be altered by the probability of
	   insertions / deletions */
	memset(Mutations1, 0, sizeof(Mutations1));
	memset(Mutations2, 0, sizeof(Mutations2));
	while (nMutations--) {
		int k1 = rand() % s->nNucleotides,
		    k2 = rand() % s->nNucleotides;
		while (Mutations1[k1] || g_InvariableSites[k1]) k1 = (k1+1) % s->nNucleotides;
		while (Mutations2[k2] || g_InvariableSites[k2]) k2 = (k2+1) % s->nNucleotides;
		Mutations1[k1] = Mutations2[k2] = 1;	
		s1->Nucleotides[k1].Value = get_nucleotide(Model, s->Nucleotides[k1].Value);
		s2->Nucleotides[k2].Value = get_nucleotide(Model, s->Nucleotides[k1].Value);
	}

	strain_delete(&s);
	strain_evolve(h, Model, s1, nGenerations, iGeneration+1);
	strain_evolve(h, Model, s2, nGenerations, iGeneration+1);
}
